#include "lzr.h"
#include "msg.h"

#define LZ_METHOD_ORIG (0)
#define LZ_METHOD_CPYN (1)
#define LZ_METHOD_RPTN (2)
#define LZ_METHOD_DICT (3)

int LZ_REPT_THRESHOLD = 3;

struct flagMgr
{
	std::vector<unsigned char> data;
	unsigned char cache;
	unsigned int  pos;
	unsigned long bits;

	flagMgr() : cache(0), pos(0), bits(-1)
	{
	}

	void add_flag(unsigned char in)
	{
		cache |= (in & 0x3) << (6 - pos);
		if(6 < (pos = pos + 2)) data.push_back(cache), cache = 0, pos = 0;
		//else pos += 2;
	}

	std::basic_string<unsigned char> get_flag()
	{
		std::basic_string<unsigned char> flg(data.begin(), data.end());
		if(pos)
		{
			flg += cache;
		}
//		bits = data.size() * 8 + pos;
		return flg;
	}

	int get_bits()
	{
		return bits = data.size() * 8 + pos;
	}
};

#define BYTE_MAX (0x7F) /*Original GUST Implementation, */
// #define BYTE_MAX (0xFF) /*It just, why not?*/

void save_hangup(std::queue<unsigned char> &hangup, flagMgr *flag, std::basic_string<unsigned char> &out)//checked, work well 
{
	if (hangup.size()) debug("Save hangup, size: %zu\n", hangup.size());

	while (hangup.size() > BYTE_MAX)   //store the part the out of range (CPYN uses 1 byte for length, but it is signed.
	{
		int n = BYTE_MAX;
		flag->add_flag(LZ_METHOD_CPYN);
		out += (unsigned char)n;
		while (n--)
		{
			out += hangup.front();
			hangup.pop();
		}
	}

	if (hangup.size() > LZ_CPYN_THRESHOLD)
	{
		flag->add_flag(LZ_METHOD_CPYN);
		out += (byte_t)hangup.size();
		while (!hangup.empty())
		{
			out += hangup.front();
			hangup.pop();
		}
	}
	else
	{
		while (!hangup.empty())
		{
			flag->add_flag(LZ_METHOD_ORIG);
			out += hangup.front();
			hangup.pop();
		}
	}
}

struct SymbDesc{
	U16 p:12;
	U16 n:4;
	I16 fast_forward;
};

SymbDesc seek_symb(byte_t *window, byte_t *buffer, byte_t *end);

int check_rpt(byte_t *buffer, byte_t *end)
{
	byte_t ch = *buffer;
	end = MIN(end, buffer + BYTE_MAX);
	int length = 0;
	while (buffer < end && *buffer++ == ch) ++length;
	return length;
}

/**
 * @brief Save repeat segment
 * @param hangup 
 * @param fm 
 * @param lzdata 
 * @param LZ_BUFFER 
 * @param LZ_DATA_END 
 * @return How many bytes processed.
 */
int record_rpt(std::queue<unsigned char> &hangup, flagMgr &fm, std::basic_string<unsigned char> &lzdata, byte_t *&LZ_BUFFER, byte_t *LZ_DATA_END, int rpt_limit = BYTE_MAX);

SymbDesc recheck_dict(SymbDesc ori_index, byte_t *buffer, byte_t *end);

unsigned int lzr(byte_t *data, long length, byte_t ** result)
{
	debug(L"lzr(): get data, length:%ld\n", length);
	bool flag_end = false;
	
	std::basic_string<byte_t> lzdata;
	std::queue<byte_t> hangup;
	flagMgr fm;
	
	byte_t *LZ_BUFFER   = data;                   //set buffer size
	byte_t *LZ_DATA_END = data + length;          //it marks the termination, shouldn't be accessed!
	
	// LZ_WINDOW ---------------- LZ_BUFFER ----- LZ_BUFFER_END
	#define LZ_BUFFER_END (LZ_BUFFER + LZ_BUFFER_SIZE) // buffer end pointer
	#define LZ_WINDOW     (LZ_BUFFER - LZ_WINDOW_SIZE)      // window pointer
	//#define REPO if(!((int)LZ_BUFFER & 0xFF)) {printf("\rProgress: %06lld/%06ldbyte(s) compressed", LZ_BUFFER - data, length);}else{}
	long long last_repo = clock();
	
	//while(LZ_WINDOW < data)//below phase: the window start at data, not LZ_WINDOW
	while (LZ_BUFFER < LZ_DATA_END)
	{
		if(clock() - last_repo >= REPO_GAP) msg(MSG_COMPRESS_PROGRESS, LZ_BUFFER - data, length), last_repo = clock();
		
		if(LZ_BUFFER == LZ_DATA_END) { flag_end = true; break; }//if process is over
		SymbDesc sdesc = seek_symb(MAX(data, LZ_WINDOW), LZ_BUFFER, MIN(LZ_BUFFER_END + LZ_BUFFER_SIZE, LZ_DATA_END));//find symbol that could match
		// the indexed dicationary is too long for buffer, maybe some unmatched, recheck it after!
		int recheck = 0;
		if (sdesc.fast_forward && sdesc.n + sdesc.fast_forward >= 2 * LZ_BUFFER_SIZE)
		{
			recheck = 1;
		}

		// fast forward process: next matched dictionary location
		int rptlen;
		while (sdesc.fast_forward > 0)
		{
			// check if repeat is more efficient
			rptlen = check_rpt(LZ_BUFFER, LZ_DATA_END);
			if (rptlen >= LZ_REPT_THRESHOLD)
			{
				// may break dictionary ref
				if (rptlen > sdesc.fast_forward)
				{
					// calc: how many bytes does it exceeded
					int overlength = rptlen - sdesc.fast_forward;
					// large enough to cover the dict
					if (sdesc.n - overlength <= 1)
					{
						sdesc.fast_forward -= record_rpt(hangup, fm, lzdata, LZ_BUFFER, LZ_DATA_END);
					}
					// too short, should not disturb the dict
					else
					{
						// too short to record, record these as normal bytes, and go ahead to next phase.
						if (sdesc.fast_forward < LZ_REPT_THRESHOLD)
						{
							while (sdesc.fast_forward--)
							{
								hangup.push(*LZ_BUFFER++);
							}
							break;
						}
						// can repeat up to fast_forward(before dict
						else sdesc.fast_forward -= record_rpt(hangup, fm, lzdata, LZ_BUFFER, LZ_DATA_END, sdesc.fast_forward);
					}
				}
				// safe repeat, just do it
				else
					sdesc.fast_forward -= record_rpt(hangup, fm, lzdata, LZ_BUFFER, LZ_DATA_END);
				continue;
			}

			//debug("ori:%02hhX\n", *LZ_BUFFER);
			hangup.push(*LZ_BUFFER++);
			sdesc.fast_forward--;
		}
		// repeat fucked dictionary
		if (sdesc.fast_forward < 0) continue;
		// repeat is more efficient!
		rptlen = check_rpt(LZ_BUFFER, MIN(LZ_BUFFER_END, LZ_DATA_END));
		if (rptlen > LZ_REPT_THRESHOLD && rptlen >= sdesc.n)//if find repeat more than dict
		{
			record_rpt(hangup, fm, lzdata, LZ_BUFFER, LZ_DATA_END);
			continue;
		}
		if (recheck) sdesc = recheck_dict(sdesc, LZ_BUFFER, MIN(LZ_DATA_END, LZ_BUFFER_END));

		if(sdesc.n > LZ_DICT_THRESHOLD)
		{
			save_hangup(hangup, &fm, lzdata); //save symbol(s) that was hung up before
			debug(L"dict: p:%d, n:%d\n", sdesc.p, sdesc.n);
			fm.add_flag(LZ_METHOD_DICT);
			byte_t temp = sdesc.p >> 4;       //build double-byte forward pointer(12 bits) et length(4 bits) 
			lzdata += temp;
			temp = (sdesc.p & 0xF) << 4 | (sdesc.n & 0xF);
			lzdata += temp;
			LZ_BUFFER += sdesc.n;//move on
		}
		else
		{
			//debug("ori:%02hhX\n", *LZ_BUFFER);
			hangup.push(*LZ_BUFFER);//temporary hang up for more information to analyze the strategy
			++LZ_BUFFER;
		}
	}
	if(!hangup.empty()) save_hangup(hangup, &fm, lzdata);
	
	LZRHD hd;
	hd.data_leng = lzdata.size();
	hd.flag_bits = fm.get_bits();
	hd.orig_leng = length;
	memcpy(hd.symb, "LZR", sizeof("LZR"));
	
	int flag_leng = (hd.flag_bits + 7)/8;
	
	byte_t *rst = new byte_t[sizeof(LZRHD) + flag_leng + hd.data_leng];
	memcpy(rst, &hd, sizeof(LZRHD));
	memcpy(rst + sizeof(LZRHD), fm.get_flag().c_str(), flag_leng);
	memcpy(rst + sizeof(LZRHD) + flag_leng, lzdata.c_str(), lzdata.size());
	
	*result = rst;
	
	return sizeof(LZRHD) + (hd.flag_bits + 7)/8 + hd.data_leng;
	
	#undef LZ_BUFFER_END 
#undef LZ_WINDOW
	#undef REPO
}

int record_rpt(std::queue<unsigned char> &hangup, flagMgr &fm, std::basic_string<unsigned char> &lzdata, byte_t *&LZ_BUFFER, byte_t *LZ_DATA_END, int rpt_limit)
{
	save_hangup(hangup, &fm, lzdata);
	byte_t st = *LZ_BUFFER;//set repeat symbol
	I8 rpt_cnt = 0;        //to count repeats
	while (st == *LZ_BUFFER && LZ_BUFFER < LZ_DATA_END && rpt_cnt < rpt_limit)
	{
		++rpt_cnt, ++LZ_BUFFER;
	}
	lzdata += (byte_t)rpt_cnt;   //append repeat account(n)
	lzdata += st;                 //append data
	fm.add_flag(LZ_METHOD_RPTN);  //set flag
	debug(L"Repeat %02hhX %d times.\n", st, rpt_cnt);
	return rpt_cnt;
}

SymbDesc seek_symb(byte_t *window, byte_t *buffer, byte_t *end)
{
	// find the nearest (least fast forward), longest segment in the
	// current buffer. meanwhile, avoid to currupt the longest segment.
	int il = end - buffer, jl = buffer - window;
	static U8 dp[(LZ_BUFFER_SIZE + LZ_BUFFER_SIZE + 1)][LZ_WINDOW_SIZE + 1];
	SymbDesc ans{ 0, 0, 0 };
	for (int j = 0; j <= jl; ++j)
	{
		if (window[j] == buffer[0]) dp[0][j] = 1;
		else dp[0][j] = 0;
	}

	for (int i = 1; i < il; ++i)
	{
		dp[i][0] = window[0] == buffer[i] ? 1 : 0;
		for (int j = 1; j < jl; ++j)
		{
			if (window[j] == buffer[i])
			{
				dp[i][j] = dp[i - 1][j - 1] + 1;
				// clamp value [0,15] (max length: 15)
				// -- do we need it?
				if (dp[i][j] > 15) dp[i][j] = 15;
				if (dp[i][j] >= ans.n)
				{
					// beginning of the matched sequence
					int beginning = j - dp[i][j] + 1;
					if (jl - beginning + ans.fast_forward > LZ_WINDOW_SIZE)
					{
						continue;
					}

					ans.n = dp[i][j];
					ans.fast_forward = i - dp[i][j] + 1;
					ans.p = jl - beginning + ans.fast_forward;
					assert(ans.fast_forward >= 0);
					assert(ans.p >= 0);
				}
			}
			else
				dp[i][j] = 0;
		}
	}

	//printf("   ");
	//for (int j = 0; j < jl; ++j)
	//{
	//	printf("%02hhx ", window[j]);
	//}
	//printf("\n");
	//for (int i = 0; i < il; ++i)
	//{
	//	printf("%02hhx:", buffer[i]);
	//	for (int j = 0; j < jl; ++j)
	//	{
	//		printf("%2d ", dp[i][j]);
	//	}
	//	printf("\n");
	//}

	// skipped too small, even found some string it is too small to be used.
	while (ans.fast_forward > LZ_DICT_THRESHOLD)
	{
		SymbDesc second_ans{ 0, 0, 0 };
		for (int i = ans.fast_forward - 1; i >= 0; --i)
			for (int j = jl - 1; j >= i; --j)
			{
				if (dp[i][j] >= second_ans.n)
				{
					if (jl - (j - dp[i][j] + 1) + second_ans.fast_forward > LZ_WINDOW_SIZE)
					{
						continue;
					}

					second_ans.n = dp[i][j];
					second_ans.fast_forward = i - dp[i][j] + 1;
					second_ans.p = jl - (j - dp[i][j] + 1) + second_ans.fast_forward;
					assert(second_ans.fast_forward >= 0);
					assert(second_ans.p >= 0);
				}
			}

		if (second_ans.n) ans = second_ans;
		// found nothing, do not update, terminate the loop immediately.
		else break;
	}

	//debug("%d %d %d\n", ans.p, ans.n, ans.fast_forward);
	return ans;
}


SymbDesc recheck_dict(SymbDesc ori_index, byte_t *buffer, byte_t *end)
{
	assert(ori_index.fast_forward == 0);
	debug(L"!!!!!!!!!!!!!Recheck!!!!!!!!!!!!!\n");
	// recover incomplete comprasion
	byte_t *index_start = buffer - ori_index.p + ori_index.n;
	buffer += ori_index.n;
	while (*buffer++ == *index_start++ && buffer < end)
	{
		ori_index.n += 1;
	}
	return ori_index;
}

//SymbDesc seek_symb(byte_t *window, byte_t *buffer, byte_t *end)
//{
//	//window               buffer  end
//	// |--------------------|-------|
//	//  <------jl----------> <-il-->
//	//  ^j                   ^i       - init
//	//           ^j                ^i - match
//	// i is length, j is location + length
//	//        p = jl - (j - i), n = i
//	int il = end - buffer, jl = buffer - window;
//	
//	SymbDesc result{ 0, 0, 0 };
//	for (int j = 0; j < jl; ++j)
//	{
//		int length = 0;
//		for (int i = 0; i < il; ++i)
//		{
//			// if window is end, or if it is mismatch, i is the length of substring
//			if (j + length >= jl || buffer[i] != window[j + length])
//			{
//				if (i >= result.n)
//				{
//					result.p = jl - j;
//					result.n = length;
//				}
//				break;
//			}
//			else
//			{
//				++length;
//			}
//		}
//	}
//	if (!result.n)
//		result.p = 0;
//
//	return result;
//}

//dustbin
/*
		{
			cmp_ary[i][j] = cmp_ary[i - 1][j - 1] + 1;
			if(lest < cmp_ary[i][j])
			{
				y = j - cmp_ary[i][j] + 1;
				lest = cmp_ary[i][j];
			}
		}
*/
